from distutils.core import setup

setup(
    name='py_amoeba',
    version='0.2.2',
    author='Chris Mutel',
    author_email='cmutel@gmail.com',
    url='https://bitbucket.org/cmutel/py-amoeba',
    packages=['py_amoeba'],
    license='BSD 2-clause; LICENSE.txt',
    long_description=open('README').read(),
)
